import {
  __publicField
} from "../../chunk-F3FYYIAV.js";
import { createRouteExpression, middlewareCreator, removeSlashes } from "../../utils";
class FastHTML {
  /**
   * @param options  The options to construct the class.
   * @throws If `ignoredRoutes` is empty and `useAppRoutes` is set to `false`.
   */
  constructor({ inject, ...options }) {
    /**
     * The service that serves a file.
     */
    __publicField(this, "_sendFile");
    /**
     * The application envent bus, to listen and get the list of "controlled routes" after
     * the application is started.
     */
    __publicField(this, "_events");
    /**
     * A function to get a possible {@link HTMLGenerator}. This is injected as a "getter"
     * to not interrupt the DIC "lifecycle": middlewares are initialized right when the
     * app starts, and injecting a reference would force the service to be initialized
     * too, even if a request is not being made.
     */
    __publicField(this, "_getHTMLGenerator");
    /**
     * The customization options for the middleware.
     */
    __publicField(this, "_options");
    /**
     * Whether or not the file is ready to be served. In case the middleware uses an
     * {@link HTMLGenerator} service, the file needs to be generated before being available,
     * and that's why this flag exists.
     */
    __publicField(this, "_fileReady", false);
    /**
     * A list of regular expression that match the routes controlled by the application.
     * This is in case the `useAppRoutes` option is set to `true`; when the application gets
     * started, an event listener will obtain all the top controlled routes, create regular
     * expressions, and save them on this property.
     */
    __publicField(this, "_routeExpressions", []);
    this._sendFile = inject.sendFile;
    this._events = inject.events;
    this._getHTMLGenerator = inject.getHTMLGenerator || (() => void 0);
    this._options = {
      file: "index.html",
      ignoredRoutes: [/\.ico$/i],
      useAppRoutes: true,
      ...options
    };
    if (!this._options.ignoredRoutes.length && !this._options.useAppRoutes) {
      throw new Error("You must provide either `ignoredRoutes` or `useAppRoutes`");
    }
    if (this._options.useAppRoutes) {
      this._setupEvents();
    }
  }
  /**
   * Generates the middleware that serves the HTML file.
   */
  getMiddleware() {
    return async (req, res, next) => {
      if (this._shouldIgnoreRoute(req.originalUrl)) {
        next();
        return;
      }
      if (this._fileReady) {
        this._sendResponse(res, next);
        return;
      }
      const htmlGenerator = this._getHTMLGenerator();
      if (!htmlGenerator) {
        this._fileReady = true;
        this._sendResponse(res, next);
        return;
      }
      try {
        await htmlGenerator.whenReady();
        this._options.file = htmlGenerator.options.file;
        this._fileReady = true;
        this._sendResponse(res, next);
      } catch (error) {
        next(error);
      }
    };
  }
  /**
   * The customization options.
   */
  get options() {
    return { ...this._options };
  }
  /**
   * Serves the HTML file to the response.
   *
   * @param res   The response object generated by the application.
   * @param next  The function to call the next middleware.
   */
  _sendResponse(res, next) {
    res.setHeader("Content-Type", "text/html");
    this._sendFile({
      res,
      next,
      filepath: this._options.file
    });
  }
  /**
   * Adds the event listener that obtains the list of "controlled routes" when
   * `useAppRoutes` is set to `true`.
   */
  _setupEvents() {
    this._events.once("afterStart", ({ app }) => {
      const [routeExpressions] = app.routes.reduce(
        (acc, route) => {
          const [expressions, processed] = acc;
          const clean = removeSlashes(route).trim();
          if (!clean || processed.includes(clean))
            return acc;
          expressions.push(createRouteExpression(clean));
          processed.push(clean);
          return acc;
        },
        [[], []]
      );
      this._routeExpressions.push(...routeExpressions);
    });
  }
  /**
   * Validates whether a route should be ignored or not. The method checks first against
   * the `ignore` option, and then against the list of "controlled routes" if
   * `useAppRoutes` is set to `true`.
   *
   * @param route  The route to validate.
   */
  _shouldIgnoreRoute(route) {
    return this._options.ignoredRoutes.some((expression) => route.match(expression)) || this._routeExpressions.some((expression) => route.match(expression));
  }
}
const fastHTMLMiddleware = middlewareCreator(
  (options = {}) => (app) => {
    const { htmlGeneratorServiceName = "htmlGenerator", ...rest } = options;
    return new FastHTML({
      inject: {
        events: app.get("events"),
        sendFile: app.get("sendFile"),
        getHTMLGenerator: () => app.try(htmlGeneratorServiceName)
      },
      ...rest
    }).getMiddleware();
  }
);
export {
  FastHTML,
  fastHTMLMiddleware
};
//# sourceMappingURL=fastHTML.js.map