import {
  __publicField
} from "../../chunk-F3FYYIAV.js";
import { middlewareCreator } from "../../utils";
class ShowHTML {
  /**
   * @param options  The options to construct the class.
   */
  constructor({ inject, ...options }) {
    /**
     * The service that serves a file.
     */
    __publicField(this, "_sendFile");
    /**
     * A function to get a possible {@link HTMLGenerator}. This is injected as a "getter"
     * to not interrupt the DIC "lifecycle": middlewares are initialized right when the
     * app starts, and injecting a reference would force the service to be initialized
     * too, even if a request is not being made.
     */
    __publicField(this, "_getHTMLGenerator");
    /**
     * The customization options for the middleware.
     */
    __publicField(this, "_options");
    /**
     * Whether or not the file is ready to be served. In case the middleware uses an
     * {@link HTMLGenerator} service, the file needs to be generated before being available,
     * and that's why this flag exists.
     */
    __publicField(this, "_fileReady", false);
    this._sendFile = inject.sendFile;
    this._getHTMLGenerator = inject.getHTMLGenerator || (() => void 0);
    this._options = {
      file: "index.html",
      ...options
    };
  }
  /**
   * Generates the middleware that serves the HTML file.
   */
  getMiddleware() {
    return async (_, res, next) => {
      if (this._fileReady) {
        this._sendResponse(res, next);
        return;
      }
      const htmlGenerator = this._getHTMLGenerator();
      if (!htmlGenerator) {
        this._fileReady = true;
        this._sendResponse(res, next);
        return;
      }
      try {
        await htmlGenerator.whenReady();
        this._options.file = htmlGenerator.options.file;
        this._fileReady = true;
        this._sendResponse(res, next);
      } catch (error) {
        next(error);
      }
    };
  }
  /**
   * The customization options.
   */
  get options() {
    return { ...this._options };
  }
  /**
   * Serves the HTML file to the response.
   *
   * @param res   The response object generated by the application.
   * @param next  The function to call the next middleware.
   */
  _sendResponse(res, next) {
    res.setHeader("Content-Type", "text/html");
    this._sendFile({
      res,
      next,
      filepath: this._options.file
    });
  }
}
const showHTMLMiddleware = middlewareCreator(
  (options = {}) => (app) => {
    const { htmlGeneratorServiceName = "htmlGenerator", ...rest } = options;
    return new ShowHTML({
      inject: {
        sendFile: app.get("sendFile"),
        getHTMLGenerator: () => app.try(htmlGeneratorServiceName)
      },
      ...rest
    }).getMiddleware();
  }
);
export {
  ShowHTML,
  showHTMLMiddleware
};
//# sourceMappingURL=showHTML.js.map