import {
  __publicField
} from "../../chunk-F3FYYIAV.js";
import { provider, statuses as statusesFn } from "../../utils";
class AppError extends Error {
  /**
   * @param message   The message of the error.
   * @param context   The context information, for the error handler.
   * @param statuses  A reference to the service that generates HTTP status codes. This
   *                  is in case the implementation wants to use a special version from
   *                  the container; otherwise, it will use the `statuses` library
   *                  directly.
   */
  constructor(message, context = {}, statuses = statusesFn) {
    super(message);
    /**
     * The date of when the error was generated.
     */
    __publicField(this, "_date");
    /**
     * The context information that can be provided to the error, and picked by the error
     * handler.
     */
    __publicField(this, "_context");
    /**
     * The service that generates HTTP status codes.
     */
    __publicField(this, "_statuses");
    this.name = this.constructor.name;
    this._date = /* @__PURE__ */ new Date();
    this._statuses = statuses;
    this._context = this._parseContext(context);
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, this.constructor);
    }
  }
  /**
   * Gets an object that can be included in a response from the application. This method
   * is a helper for the `response` getter, as it allows for the assertion of the response
   * type.
   *
   * @template T  The type of the response.
   */
  getResponse() {
    return this.response;
  }
  /**
   * Information about the error that can be included in a response. This is set using the
   * `response` key in the `context` option.
   */
  get response() {
    return this._context.response || {};
  }
  /**
   * An HTTP status code related to the error. This is set using the `status` key on the
   * `context`.
   */
  get status() {
    return this._context.status;
  }
  /**
   * Context information related to the error.
   */
  get context() {
    return this._context;
  }
  /**
   * The date of when the error was generated.
   */
  get date() {
    return this._date;
  }
  /**
   * Utility method that formats the context before saving it in the instance:
   * - If the context includes a `status` as a `string`, it will try to replace it with
   * its actual status code from the `statuses `service.
   *
   * @param context  The original context sent to the constructor.
   */
  _parseContext(context) {
    const result = { ...context };
    if (result.status && typeof result.status === "string") {
      result.status = this._statuses.code[result.status.toLowerCase()] || result.status;
    }
    return result;
  }
}
const createAppError = (...args) => new AppError(...args);
const appErrorProvider = provider((app) => {
  app.set("AppError", () => AppError);
  app.set("appError", () => createAppError);
});
export {
  AppError,
  appErrorProvider,
  createAppError
};
//# sourceMappingURL=appError.js.map