import {
  __publicField
} from "../../chunk-F3FYYIAV.js";
import urijs from "urijs";
import { deepAssignWithOverwrite } from "@homer0/deep-assign";
import { provider } from "../../utils";
class HTTP {
  /**
   * @param options  The options to construct the class.
   */
  constructor({ inject: { logger, nodeFetch }, ...options }) {
    /**
     * The service used to log information in the terminal.
     */
    __publicField(this, "_logger");
    /**
     * The node-fetch library. Since it's an ESM only module, Jimpex loads it on boot and makes
     * it available on the container.
     */
    __publicField(this, "_nodeFetch");
    /**
     * The service customization options.
     */
    __publicField(this, "_options");
    this._logger = logger;
    this._nodeFetch = nodeFetch;
    this._options = deepAssignWithOverwrite(
      {
        logRequests: false
      },
      options
    );
    this.fetch = this.fetch.bind(this);
  }
  /**
   * Makes a fetch request.
   *
   * @param url      The URL to fetch.
   * @param options  The custom options for the request.
   */
  async fetch(url, options = {}) {
    let useURL = url;
    if (options.qs) {
      useURL = urijs(url).query(options.qs).toString();
    }
    const fetchOptions = {
      method: (options.method || "get").toUpperCase(),
      body: options.body
    };
    let defaultHeaders;
    if (options.req) {
      defaultHeaders = this.getCustomHeadersFromRequest(options.req);
      const ip = this.getIPFromRequest(options.req);
      if (ip) {
        defaultHeaders["x-forwarded-for"] = ip;
      }
    }
    const headers = {
      ...defaultHeaders,
      ...options.headers
    };
    if (Object.keys(headers).length) {
      fetchOptions.headers = this.normalizeHeaders(headers);
    }
    const { logRequests } = this._options;
    if (logRequests) {
      this._logRequest(useURL, fetchOptions);
    }
    const response = await this._nodeFetch(useURL, fetchOptions);
    if (logRequests) {
      this._logResponse(response);
    }
    return response;
  }
  /**
   * Tries to get the IP address from a given request.
   *
   * @param req  The request from which it will try to obtain the IP address.
   */
  getIPFromRequest(req) {
    const headerValue = req.headers["x-forwarded-for"];
    if (headerValue) {
      return String(headerValue);
    }
    return req?.connection?.remoteAddress || req?.socket?.remoteAddress;
  }
  /**
   * Creates a dictionary with all the custom headers a request has. By custom header it
   * means all the headers which name start with `x-`.
   *
   * @param req      The request from which it will try to get the headers.
   * @param options  The options to customize the behavior with certain headers.
   */
  getCustomHeadersFromRequest(req, options = {}) {
    const { includeXForwardedHeaders = false } = options;
    return Object.keys(req.headers).reduce((acc, headerName) => {
      if (headerName.startsWith("x-") && (includeXForwardedHeaders || !headerName.startsWith("x-forwarded-"))) {
        acc[headerName] = req.headers[headerName];
      }
      return acc;
    }, {});
  }
  /**
   * It takes a dictionary of headers and normalize the names so each word will start with
   * an upper case character. This is helpful in case you added custom headers and didn't
   * care about the casing, or when copying headers from a server request, as they all
   * come tranformed into lower case.
   *
   * @param headers  The dictionary of headers to normalize.
   */
  normalizeHeaders(headers) {
    return Object.keys(headers).reduce((acc, name) => {
      const newName = name.split("-").map((part) => part.replace(/^(\w)/, (_, letter) => letter.toUpperCase())).join("-");
      acc[newName] = headers[name];
      return acc;
    }, {});
  }
  /**
   * The customization options.
   */
  get options() {
    return deepAssignWithOverwrite({}, this._options);
  }
  /**
   * Logs a request information into the terminal.
   *
   * @param url      The request URL.
   * @param options  The options for the request.
   */
  _logRequest(url, options) {
    const { method, headers } = options;
    const prefix = "REQUEST> ";
    const lines = ["--->>", `${prefix}${method} ${url}`];
    if (headers) {
      Object.keys(headers).forEach((header) => {
        const value = headers[header];
        lines.push(`${prefix}${header}: ${value}`);
      });
    }
    if (options.body) {
      lines.push(`${prefix}body: "${options.body}"`);
    }
    this._logger.info(lines);
  }
  /**
   * Logs a response information into the terminal.
   *
   * @param response  The response to log.
   */
  _logResponse(response) {
    const prefix = "RESPONSE> ";
    const lines = [
      "<<---",
      `${prefix}${response.url}`,
      `${prefix}status: ${response.status}`
    ];
    response.headers.forEach((value, header) => {
      lines.push(`${prefix}${header}: ${value}`);
    });
    this._logger.info(lines);
  }
}
const httpProvider = provider((app) => {
  app.set("http", () => {
    const config = app.getConfig();
    const logRequests = config.get("debug.logRequests") === true;
    return new HTTP({
      inject: {
        logger: app.get("logger"),
        nodeFetch: app.get("node-fetch")
      },
      logRequests
    });
  });
});
export {
  HTTP,
  httpProvider
};
//# sourceMappingURL=http.js.map