import {
  __publicField
} from "../../chunk-F3FYYIAV.js";
import { provider } from "../../utils";
const DEFAULT_CLOSE_DELAY_FOR_POST_MESSAGE = 700;
class ResponsesBuilder {
  /**
   * @param options  The options to construct the service.
   */
  constructor({ inject: { config, statuses } }) {
    /**
     * The application configuration, to get the `version` and the `postMessagesPrefix`.
     */
    __publicField(this, "_config");
    /**
     * The uility service to get HTTP status codes.
     */
    __publicField(this, "_statuses");
    this._config = config;
    this._statuses = statuses;
  }
  /**
   * Generates and send an HTML response that emits a post message.
   * The post message will be prefixed with the value of the configuration setting
   * `postMessagesPrefix`.
   *
   * @param options  The options to customize how the HTML is generated.
   */
  htmlPostMessage(options) {
    const {
      res,
      title,
      message,
      status,
      target = "window.opener",
      close = true,
      closeDelay = DEFAULT_CLOSE_DELAY_FOR_POST_MESSAGE
    } = options;
    const prefix = this._config.get("postMessagesPrefix") ?? "";
    const closeCode = close ? `setTimeout(function() { window.close(); }, ${closeDelay});` : "";
    const html = this._htmlTemplate(
      title,
      `
      (function() {
        if (${target}) {
          ${target}.postMessage('${prefix}${message}', '*');
          ${closeCode}
        }
      })();
      `
    );
    const useStatus = typeof status === "undefined" ? this._statuses("ok") : this._normalizeStatus(status);
    res.setHeader("Content-Type", "text/html");
    res.setHeader("Cache-Control", "no-cache, max-age=0, must-revalidate, no-store");
    res.status(useStatus);
    res.write(html);
    res.end();
  }
  /**
   * Generates and sends a JSON response.
   *
   * @param options  The options to customize how the JSON is generated.
   * @example
   *
   * <caption>The generated looks like this.</caption>
   *
   * {
   *   metadata: {
   *     version: 'dev',
   *     status: 200,
   *   },
   *   data: { foo: 'bar' },
   * }
   *
   */
  json(options) {
    const { res, data, status, metadata = {} } = options;
    const useStatus = typeof status === "undefined" ? this._statuses("ok") : this._normalizeStatus(status);
    res.status(useStatus);
    res.json({
      metadata: {
        version: this._config.get("version"),
        status: useStatus,
        ...metadata
      },
      data
    });
    res.end();
  }
  /**
   * Generates a basic HTML template for the service to use when generating a post message
   * response.
   *
   * @param title  The HTML `<title />` attribute.
   * @param code   Javascript code to be wrapped on a `<script />` tag.
   */
  _htmlTemplate(title, code) {
    return `
      <!DOCTYPE html>
      <html>
        <head>
          <meta charset="utf-8" />
          <meta http-equiv="x-ua-compatible" content="ie=edge" />
          <meta name="viewport" content="width=device-width, initial-scale=1" />
          <title>${title}</title>
        </head>
        <body>
          <script type="text/javascript">
          ${code}
        </script>
      </body>
    </html>
    `;
  }
  /**
   * Utility method used to make sure a recevied status is a valid status code. If the
   * status is a string, the method will try to find the code from the `statuses` package.
   *
   * @param status  The status to normalize.
   * @returns If `status` is a string, but there's no valid code, it will return 200.
   */
  _normalizeStatus(status) {
    let useStatus;
    try {
      if (typeof status === "string") {
        useStatus = this._statuses(status);
      } else {
        this._statuses(status);
        useStatus = status;
      }
    } catch (_) {
      useStatus = this._statuses("ok");
    }
    return useStatus;
  }
}
const responsesBuilderProvider = provider((app) => {
  app.set(
    "responsesBuilder",
    () => new ResponsesBuilder({
      inject: {
        config: app.getConfig(),
        statuses: app.get("statuses")
      }
    })
  );
});
export {
  ResponsesBuilder,
  responsesBuilderProvider
};
//# sourceMappingURL=responsesBuilder.js.map